@@ -5,6 +5,7 @@ from typing import List
 import aiohttp
 from aiohttp.helpers import BasicAuth
 import websocket
+import websockets
 from datetime import datetime
 from pymee.const import DeviceApp, DeviceOS, DeviceType
 from pymee.model import HomeeNode
@@ -22,6 +23,7 @@ class Homee:
         reconnectInterval: int = 5000,
         reconnect: bool = True,
         maxRetries: int = 5,
+        loop: asyncio.AbstractEventLoop = None,
     ) -> None:
 
         self.host = host
@@ -38,13 +40,15 @@ class Homee:
         self.nodes: List[HomeeNode] = []
         self.groups = []
         self.relationships = []
-        self.ws = None
         self.token = ""
         self.expires = 0
         self.connected = False
         self.retries = 0
         self.shouldClose = False
 
+        self._message_queue = asyncio.Queue(loop=loop)
+        self._connected_event = asyncio.Event(loop=loop)
+
     async def get_access_token(self):
         """Asynchronously attempts to get an access token from the homee host using username and password."""
 
@@ -92,7 +96,7 @@ class Homee:
 
         except:
             await client.close()
-            raise AuthenticationFailedException()
+            raise AuthenticationFailedException
 
         await client.close()
         return self.token
@@ -105,15 +109,15 @@ class Homee:
         try:
             await self.get_access_token()
         except:
-            raise Exception("Unable to connect. Authentication failed.")
+            raise AuthenticationFailedException
 
-        self.open_ws()
+        await self.open_ws()
 
     def start(self):
         """Wraps run() with asyncio.create_task() and returns the resulting task."""
         return asyncio.create_task(self.run())
 
-    def open_ws(self):
+    async def open_ws(self):
         """Opens the websocket connection assuming an access token was already received. Runs until connection is closed again."""
 
         self._log("Opening websocket...")
@@ -127,20 +131,37 @@ class Homee:
             return
 
         try:
-            self.ws = websocket.WebSocketApp(
-                url=f"{self.ws_url}/connection?access_token={self.token}",
+            async with websockets.connect(
+                uri=f"{self.ws_url}/connection?access_token={self.token}",
                 subprotocols=["v2"],
-                on_open=self._ws_on_open,
-                on_message=self._ws_on_message,
-                on_close=self._ws_on_close,
-                on_error=self._ws_on_error,
-            )
-            self.ws.run_forever(origin=self.url)
-        except:
+            ) as ws:
+                await self._ws_on_open()
+                while not self.shouldClose:
+                    receive_task = asyncio.ensure_future(self._ws_receive_handler(ws))
+                    send_task = asyncio.ensure_future(self._ws_send_handler(ws))
+                    done, pending = await asyncio.wait(
+                        [receive_task, send_task],
+                        return_when=asyncio.FIRST_COMPLETED,
+                    )
+                    for task in pending:
+                        task.cancel()
+        except Exception as e:
             # TODO retry logic
-            raise Exception("Unable to connect due to a websocket error")
+            self._ws_on_error(e)
+            raise e
 
-    def _ws_on_open(self):
+        self._ws_on_close()
+
+    async def _ws_receive_handler(self, ws: websockets.WebSocketClientProtocol):
+        msg = await ws.recv()
+        await self._ws_on_message(msg)
+
+    async def _ws_send_handler(self, ws: websockets.WebSocketClientProtocol):
+        msg = await self._message_queue.get()
+        if self.connected and not self.shouldClose:
+            await ws.send(msg)
+
+    async def _ws_on_open(self):
         """Websocket on_open callback."""
 
         self._log("Connection to websocket successfull")
@@ -149,19 +170,18 @@ class Homee:
         self.retries = 1
 
         self.on_connected()
-        self.send("GET:all")
+        await self.send("GET:all")
 
-    def _ws_on_message(self, msg: str):
+    async def _ws_on_message(self, msg: str):
         """Websocket on_message callback."""
 
-        self._handle_message(json.loads(msg))
+        await self._handle_message(json.loads(msg))
 
     def _ws_on_close(self):
         """Websocket on_close callback."""
         # if not self.shouldClose and self.retries <= 1:
 
         self.connected = False
-        self.ws = None
 
         self.on_disconnected()
         # if self.shouldReconnect and not self.shouldClose:
@@ -173,13 +193,13 @@ class Homee:
         self._log(f"An error occurred: {error}")
         self.on_error(error)
 
-    def send(self, msg: str):
+    async def send(self, msg: str):
         """Send a raw string message to homee."""
 
-        if not self.connected or self.ws == None:
+        if not self.connected or self.shouldClose:
             return
 
-        self.ws.send(msg)
+        await self._message_queue.put(msg)
 
     async def reconnect(self):
         """TODO"""
@@ -192,12 +212,7 @@ class Homee:
 
         self.shouldClose = True
 
-        if self.ws != None:
-            self.ws.close()
-        self._log("Connection closed")
-        self.on_disconnected()
-
-    def _handle_message(self, msg: dict):
+    async def _handle_message(self, msg: dict):
         """Internal handleing of incoming homee messages."""
 
         msgType = None
@@ -215,6 +230,7 @@ class Homee:
             self.nodes = list(map(lambda n: HomeeNode(n), msg["all"]["nodes"]))
             self.groups = msg["all"]["groups"]
             self.relationships = msg["all"]["relationships"]
+            self._connected_event.set()
         elif msgType == "attribute":
             self._handle_attribute_change(msg["attribute"])
         elif msgType == "groups":
@@ -254,18 +270,18 @@ class Homee:
     def get_node_index(self, nodeId: int):
         return next((i for i, node in enumerate(self.nodes) if node.id == nodeId), -1)
 
-    def set_value(self, deviceId: int, attributeId: int, value: int):
+    async def set_value(self, deviceId: int, attributeId: int, value: int):
         """Set the target value of an attribute of a device."""
 
         self._log(f"Set value: Device: {deviceId} Attribute: {attributeId} To: {value}")
-        self.send(
+        await self.send(
             f"PUT:/nodes/{deviceId}/attributes/{attributeId}?target_value={value}"
         )
 
-    def play_homeegram(self, homeegramId: int):
+    async def play_homeegram(self, homeegramId: int):
         """Invoke a homeegram."""
 
-        self.send(f"PUT:homeegrams/{homeegramId}?play=1")
+        await self.send(f"PUT:homeegrams/{homeegramId}?play=1")
 
     @property
     def url(self):
@@ -279,6 +295,10 @@ class Homee:
 
         return f"ws://{self.host}:7681"
 
+    @property
+    def is_connected(self):
+        return self._connected_event.wait()
+
     def on_reconnect(self):
         """TODO"""
         pass
@@ -302,7 +322,7 @@ class Homee:
         """Called when an 'attribute' message was received and an attribute was updated. Contains the parsed json attribute data and the corresponding node instance."""
 
     def _log(self, msg: str):
-        logging.info(msg)
+        logging.debug(msg)
 
 
 class HomeeException(Exception):
@@ -310,4 +330,4 @@ class HomeeException(Exception):
 
 
 class AuthenticationFailedException(HomeeException):
-    """Raised if no valid access token could be acquired."""
\ No newline at end of file
+    """Raised if no valid access token could be acquired."""
